import numbers, re
from abc import ABCMeta, abstractmethod, abstractproperty
from restrainer.constraints.utils import is_email, data_to_string_type, is_ip, is_mac
from restrainer.constraints.exceptions import *

class Constraint(metaclass=ABCMeta):

	def __init__(self):
		pass


	'''
		property identifying constraint in scheme definition
		Ex.

		"name" : {
			"type" : "string" - 'type' is name here
		}
	'''
	@property
	def name(self):
		pass

	'''
		For most cases this should stay False. It may True when it don't matter if user data is null.
		Ex. required constraint should allow this
	'''
	@property
	def accept_null(self):
	    return False

	'''
		This method decides if validation is successful or not.

		value - passed field value
		constraint_value - value of key in scheme(key is name property)

		It shall return True or False depending on validation status.
		Although returning dictionary will result in False, because dictionary key, value pairs
		will be shown as additional info in error message.
	'''
	@abstractmethod
	def validate(self, value, constraint_value, field_name, doc):
		pass

class TypeConstraint(Constraint):

	def __init__(self):
		super(TypeConstraint, self).__init__()
		self._types = {
			"integer" : int,
			"float" : float,
			"string" : str,
			"boolean" : bool,
			"list" : list,
			"object" : dict,
			"numeric" : numbers.Number
		}
	

	def register_type(self, name, cls):
		for key, type in self._types.items():
			if name in self._types:
				raise ConstrainException("Name '" + key + "' already registered.")
		self._types[name] = cls

	def name(self):
		return "type"

	def validate(self, value, constraint_value, field_name, doc):
		for key, type in self._types.items():
			if constraint_value == key and not isinstance(value, type):
				return {"type" : key}
		return True

class ListTypeConstraint(TypeConstraint):

	def __init__(self):
		super(ListTypeConstraint, self).__init__()

	def name(self):
		return "list_type"

	def validate(self, value, constraint_value, field_name, doc):
		try:
			for item in value:
				for key, type in self._types.items():
					if constraint_value == key and not isinstance(item, type):
						return {"list_type" : constraint_value}
			return True
		except:
			return {"list_type" : constraint_value}

class ExsitanceConstraint(Constraint):

	def __init__(self):
		super(Constraint, self).__init__()

	def name(self):
		return "required"

	def accept_null(self):
	    return True

	def validate(self, value, constraint_value, field_name, doc):
		if constraint_value == True:
			return value != None
		return True


class ValueConstraint(Constraint):

	def __init__(self):
		super(ValueConstraint, self).__init__()
	
	def name(self):
		return "value"

	def validate(self, value, constraint_value, field_name, doc):
		if not value in constraint_value:
			return {"allowed" : constraint_value}
		return True
		
class MinConstraint(Constraint):

	def __init__(self):
		super(MinConstraint, self).__init__()

	def name(self):
		return "min"

	def validate(self, value, constraint_value, field_name, doc):
		min = constraint_value
		if isinstance(value, numbers.Number):
			if value < min:
				return {"min" : min}
		elif isinstance(value, str) or isinstance(value, list) or isinstance(value, dict):
			if len(value) < min:
				return {"min" : min}
		return True	

class MaxConstraint(Constraint):

	def __init__(self):
		super(MaxConstraint, self).__init__()

	def name(self):
		return "max"

	def validate(self, value, constraint_value, field_name, doc):
		max = constraint_value
		if isinstance(value, numbers.Number):
			if value > max:
				return {"max" : max}
		elif isinstance(value, str) or isinstance(value, list) or isinstance(value, dict):
			if len(value) > max:
				return {"max" : max}
		return True	

class BetweenConstraint(Constraint):

	def __init__(self):
		super(BetweenConstraint, self).__init__()

	def name(self):
		return "between"

	def validate(self, value, constraint_value, field_name, doc):
		min = constraint_value[0]
		max = constraint_value[1]

		if isinstance(value, numbers.Number):
			if value < min or value > max:
				return {"min" : min, "max" : max}
		elif isinstance(value, str) or isinstance(value, list) or isinstance(value, dict):
			if len(value) < min or len(value) > max:
				return {"min" : min, "max" : max}
		return True	

class SizeConstraint(Constraint):

	def __init__(self):
		super(SizeConstraint, self).__init__()

	def name(self):
		return "size"

	def validate(self, value, constraint_value, field_name, doc):
		if isinstance(value, numbers.Number) and value != constraint_value:
			return {"size" : constraint_value}
		elif isinstance(value, str) and len(value) != constraint_value:
			return {"size" : constraint_value}
		elif isinstance(value, list) and len(value) != constraint_value:
			return {"size" : constraint_value}
		return True

class FormatConstraint(Constraint):

	def __init__(self):
		super(FormatConstraint, self).__init__()

	def name(self):
		return "data_format"

	def validate(self, value, constraint_value, field_name, doc):

		if isinstance(value, str):

			#FIX ME HER
			if constraint_value == "email" and not is_email(value):
				return {"data_format" : constraint_value}
			elif constraint_value == "ip" and not is_ip(value):
				return {"data_format" : constraint_value}
			elif constraint_value == "mac" and not is_mac(value):
				return {"data_format" : constraint_value}

			return True

		return {"data_format" : constraint_value}	
		
		



#This one lets you pass function as a constraint_value
class ValidatorConstraint(Constraint):

	def __init__(self):
		super(ValidatorConstraint, self).__init__()
		
	def name(self):
		return "validator"

	def validate(self, value, constraint_value, field_name, doc):
		success = bool(constraint_value['function'](value))
		if not success:
			if "message" in constraint_value:
				return {"message" : constraint_value["message"]}
			return False
		return True

class RegexConstraint(Constraint):

	def __init__(self):
		super(RegexConstraint, self).__init__()

	def name(self):
		return "regex"

	def validate(self, value, constraint_value, field_name, doc):
		reg = re.compile(constraint_value)
		if value != None:
			success = (reg.match(value) != None)
			if success:
				return True
			else:
				return {"regex" : constraint_value} 
		return {"regex" : constraint_value} 


class CoerceConstraint(Constraint):

	def __init__(self):
		super(CoerceConstraint, self).__init__()

	def name(self):
		return "coerce"

	def validate(self, value, constraint_value, field_name, doc):
		if hasattr(constraint_value, '__call__'):
			try:
				doc[field_name] = constraint_value(value)
				return True		
			except:
				return False
		raise ConstrainException("Coerce must be callable.")



import re, numbers, socket


email_reg = re.compile(r"[^@]+@[^@]+\.[^@]+")
mac_reg = re.compile(r"[0-9a-f]{2}([-:])[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$")

def is_email(string):
	try:
		return re.match(email_reg, string) != None
	except:
		return False

def is_ip(string):
	try:
		print(string)
		socket.inet_aton(string)
		return True
	except socket.error:
		return False

def is_mac(string):
	if re.match(mac_reg, string.lower()):
		return True
	return False

def data_to_string_type(data):
	if isinstance(data, str):
		return "string"
	elif isinstance(data, bool):
		return "boolean"
	elif isinstance(data, numbers.Number):
		return "numeric"
	elif isinstance(data, list):
		return "list"
	elif isinstance(data, dict):
		return "dictionary"
	elif isinstance(data, None):
		return "null"
	return None


import numbers
from restrainer.constraints.constraints import *
from restrainer.constraints.exceptions import *

class Validator(object):

	def __init__(self, rules, constrainers = []):
		self.__errors = []
		self.__errors_count = 0
		self.__rules = rules
		self.__constrainers = []
		if len(constrainers) == 0:
			for default_constraint in self.__default_constrains():
				self.load_constraint(default_constraint)
		else:
			for constrainer in constrainers:
				self.load_constraint(constrainer)

		self.__type_constraint = TypeConstraint()
		self.__build_method = self.__build_error

	def __default_constrains(self):
		return [ExsitanceConstraint(), ValueConstraint(),
		 		MinConstraint(), MaxConstraint(), BetweenConstraint(),
		 		SizeConstraint(), FormatConstraint(), ListTypeConstraint(),
		 		ValidatorConstraint(), RegexConstraint(), CoerceConstraint()]

	def __find_constrainer(self, name):
		for constrainer in self.constraints():
			if name == constrainer.name():
				return constrainer
		return None

	def set_type_constraint(self, type_constraint):
		self.__type_constraint = type_constraint

	def register_type(self, name, cls):
		self.__type_constraint.register_type(name, cls)

	def set_build_method(self, method):
		self.__build_method = method

	def constraints(self):
		return self.__constrainers + [self.__type_constraint]

	def load_constraint(self, constraint):
		if isinstance(constraint, Constraint):
			for constrainer in self.__constrainers:
				if constraint.name() == constrainer.name():
					raise ConstrainException("Constrainer with name'" + constraint.name() + "'already registered.")
			self.__constrainers.append(constraint)

	def fails(self):
		return self.__errors_count > 0

	def errors(self):
		return self.__errors

	def validate(self, doc = {}, rules = None):
		if rules == None:
			rules = self.__rules
		self.__errors_count = 0
		self.__errors = []
		self.__errors += self.__validate_rules(doc, rules = rules)
		return self.__errors

	def __validate_rules(self, doc, rules = None, index = None):

		errors = []


		for field_name, constraints in rules.items():

			data_value = doc.get(field_name, None)

			for constraint, constraint_value in constraints.items():

				if constraint == "items":

					if data_value == None or not isinstance(data_value, list):
						continue

					errors_collection = {field_name : []}
					for index, item in enumerate(data_value):
						errors_collection[field_name] += self.__validate_rules(item, rules = constraint_value, index = index)
					errors.append(errors_collection)

				elif constraint == "properties":

					if data_value == None or not isinstance(data_value, dict):
						continue

					target_errors = self.__validate_rules(data_value, rules = constraint_value)
					errors.append({ field_name :  target_errors })
				
				else:

					constrainer = self.__find_constrainer(constraint)
					if constrainer == None:
						raise ConstrainException("No constrainer found for attribute: '" + constraint + "'.")

					if (constrainer.accept_null or (data_value != None and not constrainer.accept_null)):
						validation_result = constrainer.validate(data_value, constraint_value, field_name, doc)
						succeeded = bool(validation_result) and not isinstance(validation_result, dict)

						if not succeeded:
							if isinstance(validation_result, dict):
								if index != None:
									errors.append(self.__build_method(field_name, constraint, index = index, **validation_result))
								else:
									errors.append(self.__build_method(field_name, constraint, **validation_result))
							else:
								if index != None:
									errors.append(self.__build_method(field_name, constraint, index = index))
								else:
									errors.append(self.__build_method(field_name, constraint))
							self.__errors_count += 1

		return errors

	def __build_error(self, field_name, constraint, **kwargs):
		error = {"field" : field_name, "constraint" : constraint}
		for key, value in kwargs.items():
			error[key] = value
		return error

'''
PERSONAL THOUGHTS DON'T READ

	Example rules set:
		//Will go with this one
		{
			"field_1" : {
				"type" : "string",
				"max" : 20,
				"data_format" : "email",
				"required" : True
			},
			"field_2" : {
				"type" : "numeric",
				"between" : [1, 100]
			}
		}
	

	VS

	{
		"field_1" : ["type:string|max:20|data_format:email|required:True"],
		"field_2" : ["type:numeric|between:1:100"]
	}

	Example evaluated object:
	{"field_1" : "abcdefghijklmnouprstwvxz"}

	{
		"age" : {
			"type" : "numeric",
			"required" : True
		}, 
		"participants" : {
			"type" : "list",
			"items" : {
				"name" : {
					"type" : "string",
					"required" : True
				},
				"parties" : {
					"type" : "list",
					"items" : {
						"name" : {
							"required" : True,
							"type" : "string"
						},
						"support" : {
							"type" : "numeric"
						}
					}
				}
			}
		}
	}

	[
		{
			"field" : "participants",
			"items" : [
				{
					"index" : 2
					"field" : "name",
					"constraint" : "type",
					"type" : "string"
				}
			]
		}
	]

	Types:
		-numeric
		-string
		-object
		-boolean
		-object
		-list(for primitves)
		-object_list(for objects)
'''


from distutils.core import setup
setup(
  name = 'restrainer',
  packages = ['restrainer', 'restrainer/constraints'], # this must be the same as the name above
  version = '1.3',
  description = 'Lightweight, flexible validation library.',
  author = 'Dawid Macek',
  author_email = 'dawidmacek42@gmail.com',
  url = 'https://github.com/Palkovsky/Strainer', # use the URL to the github repo
  download_url = 'https://github.com/Palkovsky/Strainer/tree/master/strainer', # I'll explain this in a second
  keywords = ['validation', 'library', 'rest api'], # arbitrary keywords
  classifiers = [],
)